package com.unitesk.requality.api;

import java.net.URI;

import javax.ws.rs.core.UriBuilder;

import org.eclipse.jetty.server.Server;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.glassfish.jersey.jetty.JettyHttpContainerFactory;
import org.glassfish.jersey.media.multipart.MultiPartFeature;
import org.glassfish.jersey.server.ResourceConfig;
import org.osgi.framework.BundleContext;

import com.unitesk.requality.api.services.MiscService;
import com.unitesk.requality.api.services.document.DocumentService;
import com.unitesk.requality.api.services.nodes.NodesService;
import com.unitesk.requality.api.services.project.ProjectService;
import com.unitesk.requality.api.services.report.ReportService;
import com.unitesk.requality.api.services.report.ReportSettingsService;
import com.unitesk.requality.api.services.requirement.RequirementService;
import com.unitesk.requality.core.RLogger;

public class Activator extends AbstractUIPlugin {
  public static final String HARDCODED_KEY = "36AC0A0F7F271314B72986EB54DB2343";
  public static final String PLUGIN_ID = "com.unitesk.requality.api"; //$NON-NLS-1$
  private static Activator plugin;

  private Server httpServer;
  private BaseService.SecurityFilter securityFilter;

  public Activator() {}

  /*
   * (non-Javadoc)
   * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
   */
  public void start(BundleContext context) throws Exception {
    super.start(context);
    plugin = this;

    int port = 9988;
    URI baseUri = UriBuilder.fromUri("http://localhost").port(port).build();
    ResourceConfig config =
        new ResourceConfig()
            .registerClasses(
                DocumentService.class,
                NodesService.class,
                ProjectService.class,
                ReportService.class,
                ReportSettingsService.class,
                RequirementService.class,
                MiscService.class,
                BaseService.JsonWriter.class,
                BaseService.ExceptionHandler.class,
                MultiPartFeature.class,
                BaseService.FallbackService.class)
            .registerInstances(securityFilter = new BaseService.SecurityFilter(HARDCODED_KEY));  
    try{
      httpServer = JettyHttpContainerFactory.createServer(baseUri, config);
      RLogger.logInfo("HTTP server started on " + baseUri);
    }
    catch(Exception e){
      RLogger.logInfo("HTTP server start error " + e.getMessage());
    }
  }

  public void setApiKey(String apiKey) {
    securityFilter.setApiKey(apiKey);
    ;
  }

  /*
   * (non-Javadoc)
   * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
   */
  public void stop(BundleContext context) throws Exception {
    plugin = null;
    httpServer.stop();
    super.stop(context);
  }

  /**
   * Returns the shared instance
   *
   * @return the shared instance
   */
  public static Activator getDefault() {
    return plugin;
  }

  /**
   * Returns an image descriptor for the image file at the given plug-in relative path
   *
   * @param path the path
   * @return the image descriptor
   */
  public static ImageDescriptor getImageDescriptor(String path) {
    return imageDescriptorFromPlugin(PLUGIN_ID, path);
  }
}
