package com.unitesk.requality.api.models;

import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.unitesk.requality.core.TreeNode;

public class IDTree extends JsonOutputBase {
  private IDMapping current;
  private String type;
  private List<IDTree> children;

  public IDTree(IDMapping current, String type, List<IDTree> children) {
    this.current = current;
    this.type = type;
    this.children = children;
  }

  public IDMapping getCurrent() {
    return current;
  }

  public String getType() {
    return type;
  }

  public List<IDTree> getChildren() {
    return children;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((children == null) ? 0 : children.hashCode());
    result = prime * result + ((current == null) ? 0 : current.hashCode());
    result = prime * result + ((type == null) ? 0 : type.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    IDTree other = (IDTree) obj;
    if (children == null) {
      if (other.children != null) return false;
    } else if (!children.equals(other.children)) return false;
    if (current == null) {
      if (other.current != null) return false;
    } else if (!current.equals(other.current)) return false;
    if (type == null) {
      if (other.type != null) return false;
    } else if (!type.equals(other.type)) return false;
    return true;
  }

  @Override
  public String toString() {
    return "IDTree [current=" + current + ", type=" + type + ", children=" + children + "]";
  }

  public static IDTree fromNode(TreeNode node) {
    List<IDTree> childTrees =
        Stream.of(node.getSortedWorkChildren(null))
            .map(IDTree::fromNode)
            .collect(Collectors.toList());

    return new IDTree(IDMapping.fromNode(node), node.getType(), childTrees);
  }
}
