package com.unitesk.requality.api.models;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.EnumDefinitionValue;

public class JsonAttrDeserialize  extends JsonDeserializer<JsonAttrValue> {

  @SuppressWarnings("unchecked")
  @Override
  public JsonAttrValue deserialize(JsonParser jp, DeserializationContext ctxt)
      throws IOException, JsonProcessingException {
    Map<String, Object> map = jp.readValueAs(Map.class);
    return collectValue(null, map,false);
  }
  private JsonAttrValue collectValue(AttributeType at, Map<String,Object> map,boolean enumval){
    if(at == null){
      at = AttributeType.STRING;
      if(map.containsKey("type"))
        at = AttributeType.valueOf(map.get("type").toString());
    }
    switch(at){
      case ENUM:{
        AppliedEnumValue aev = new AppliedEnumValue();
        aev.setType(at);
        Object bk = map.get("enumName"), comm = map.get("comment");
        aev.setBaseKey(bk!=null?bk.toString():null);
        //aev.setComment(comm!=null?comm.toString():null);
        aev.setValue(map.get("value"));
        return aev;
      }
      case LIST:{
        AttributeType vt = AttributeType.STRING;
        List<Map<String,Object>> val = (List<Map<String,Object>>) map.get("value");
        if(map.containsKey("valuesType"))
          vt = AttributeType.valueOf(map.get("valuesType").toString());
        List<JsonAttrValue> java = new LinkedList<>();
        for(Map<String,Object> m :val){
          JsonAttrValue jav = collectValue(vt, m,true);
          java.add(jav);
        }
        GJsonListValue<JsonAttrValue> v = new GJsonListValue<>();
        v.setValue(java);
        v.setListType(vt);
        v.setType(at);
        return v;
      }
      case ENUM_DEFINITION:{
        AttributeType vt = AttributeType.STRING;
        List<Map<String,Object>> val = (List<Map<String,Object>>) map.get("value");
        if(map.containsKey("valuesType"))
          vt = AttributeType.valueOf(map.get("valuesType").toString());
        List<JsonEnumValue> java = new LinkedList<>();
        for(Map<String,Object> m :val){
          JsonAttrValue jav = collectValue(vt, m,true);
          java.add((JsonEnumValue) jav);
        }
        JsonEnumDefValue v = new JsonEnumDefValue();
        v.setValue(java);
        v.setListType(vt);
        v.setType(at);
        return v;
      }
      default:{
      
      }
    }
    if(enumval)
    {
      JsonEnumValue jev = new JsonEnumValue();
      Object comm = map.get("comment");
      jev.setValue(comm!=null?comm.toString():null);
      jev.setType(at);
      jev.setValue(map.get("value"));
      return jev;
    }
    Object val = map.get("value");
    JsonAttrValue jav = new JsonAttrValue();
    jav.setType(at);
    jav.setValue(val);
    return jav;
  }
}
