package com.unitesk.requality.api.models;

import javax.ws.rs.core.Response.Status;

import com.fasterxml.jackson.annotation.JsonIgnore;

public class JsonError extends JsonOutputBase {
  public static JsonError INTERNAL_ERROR = new JsonError("internal_error", "Internal server error");

  private Status httpStatus;
  private String code;
  private String message;
  private JsonOutputBase extra;

  public JsonError(String code, String message) {
    this(Status.INTERNAL_SERVER_ERROR, code, message, JsonEmpty.INSTANCE);
  }

  public JsonError(String code, String message, JsonOutputBase extra) {
    this(Status.INTERNAL_SERVER_ERROR, code, message, extra);
  }

  public JsonError(Status httpStatus, String code, String message) {
    this(httpStatus, code, message, JsonEmpty.INSTANCE);
  }

  public JsonError(Status httpStatus, String code, String message, JsonOutputBase extra) {
    this.httpStatus = httpStatus;
    this.code = code;
    this.message = message;
    this.extra = extra;
  }

  @JsonIgnore
  public Status getHttpStatus() {
    return httpStatus;
  }

  public String getCode() {
    return code;
  }

  public String getMessage() {
    return message;
  }

  public JsonOutputBase getExtra() {
    return extra;
  }

  public JsonException exception() {
    return new JsonException(this);
  }

  public void raise() throws JsonException {
    throw new JsonException(this);
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((code == null) ? 0 : code.hashCode());
    result = prime * result + ((message == null) ? 0 : message.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    JsonError other = (JsonError) obj;
    if (code == null) {
      if (other.code != null) return false;
    } else if (!code.equals(other.code)) return false;
    if (message == null) {
      if (other.message != null) return false;
    } else if (!message.equals(other.message)) return false;
    return true;
  }

  @Override
  public String toString() {
    return "Error [code=" + code + ", message=" + message + "]";
  }
}
