package com.unitesk.requality.api.models;

import java.util.Objects;

import com.fasterxml.jackson.databind.JsonNode;

public class JsonResponseWrapper {
  private JsonOutputBase object;
  private JsonNode rawJson;

  public JsonResponseWrapper(JsonOutputBase object) {
    this.object = Objects.requireNonNull(object, "object == null");
    this.rawJson = null;
  }

  public JsonResponseWrapper(JsonNode rawJson) {
    this.rawJson = Objects.requireNonNull(rawJson, "rawJson == null");
    this.object = null;
  }

  public JsonResponseWrapper(Object any) {
    if (any instanceof JsonNode) {
      this.object = null;
      this.rawJson = (JsonNode) any;
    } else if (any instanceof JsonOutputBase) {
      this.object = (JsonOutputBase) any;
      this.rawJson = null;
    } else {
      throw new IllegalArgumentException(
          "Supplied object must be either JsonNode or JsonOutputBase");
    }
  }

  public boolean hasObject() {
    return object != null;
  }

  public boolean hasError() {
    return object != null && object instanceof JsonError;
  }

  public JsonOutputBase getObject() {
    return Objects.requireNonNull(object, "object is not set");
  }

  public void setObject(JsonOutputBase object) {
    this.object = Objects.requireNonNull(object, "object == null");
    this.rawJson = null;
  }

  public boolean hasRawJson() {
    return rawJson != null;
  }

  public JsonNode getRawJson() {
    return Objects.requireNonNull(rawJson, "rawJson is not set");
  }

  public void setRawJson(JsonNode rawJson) {
    this.rawJson = Objects.requireNonNull(rawJson, "rawJson == null");
    this.object = null;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((object == null) ? 0 : object.hashCode());
    result = prime * result + ((rawJson == null) ? 0 : rawJson.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    JsonResponseWrapper other = (JsonResponseWrapper) obj;
    if (object == null) {
      if (other.object != null) return false;
    } else if (!object.equals(other.object)) return false;
    if (rawJson == null) {
      if (other.rawJson != null) return false;
    } else if (!rawJson.equals(other.rawJson)) return false;
    return true;
  }

  @Override
  public String toString() {
    return "JsonResponseWrapper [object=" + object + ", rawJson=" + rawJson + "]";
  }
}
