package com.unitesk.requality.api.services.document;

import java.io.StringWriter;
import java.util.UUID;
import java.util.stream.Collectors;

import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.core.Response.Status;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import com.fasterxml.jackson.databind.node.TextNode;
import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.Utils;
import com.unitesk.requality.api.models.IDList;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.JsonEmpty;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

import static com.unitesk.requality.api.services.project.ProjectService.getProject;

@Path("/projects/{p}/documents")
public class DocumentService extends BaseService {
  public static final JsonError DOCUMENT_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "document_not_found", "Specified document does not exists");

  @GET
  public IDList httpList(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    TreeNode docRoot = db.getNode(Document.getTypeRootQId());
    return new IDList(
        Utils.stream(new DeepFirstTreeWalker(docRoot, false, true, true).iterator())
            .filter(x -> x.getType().equals(Document.TYPE_NAME))
            .map(IDMapping::fromNode)
            .collect(Collectors.toList()));
  }

  @GET
  @Path("/{d}")
  public TextNode httpGet(@PathParam("p") String project, @PathParam("d") UUID docId)
      throws JsonException {
    Document doc = getDocument(project, docId);
    try {
      org.w3c.dom.Document xml = doc.getDOM();
      StringWriter sw = new StringWriter();
      Transformer transformer = TransformerFactory.newInstance()
          .newTransformer();
      transformer.setOutputProperty(OutputKeys.METHOD, "xml");
      transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
          transformer.transform(new DOMSource(xml), new StreamResult(sw));
      return new TextNode(sw.toString());
    } catch (Exception e) {
      e.printStackTrace();
      throw JsonError.INTERNAL_ERROR.exception();
    }
  }

  @DELETE
  @Path("/{d}")
  public JsonEmpty httpDelete(@PathParam("p") String project, @PathParam("d") UUID docId)
      throws JsonException {
    getDocument(project, docId).delete();
    return JsonEmpty.INSTANCE;
  }

  public static Document getDocument(TreeDB project, UUID docId) throws JsonException {
    TreeNode n = project.getNode(docId);

    if (n == null || !n.getType().equals(Document.TYPE_NAME)) {
      throw DOCUMENT_NOT_FOUND.exception();
    }

    return (Document) n;
  }

  public static Document getDocument(String project, UUID docId) throws JsonException {
    return getDocument(getProject(project), docId);
  }
}
