package com.unitesk.requality.api.services.nodes;

import static com.unitesk.requality.api.services.project.ProjectService.getProject;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import org.glassfish.jersey.media.multipart.FormDataBodyPart;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.FormDataParam;
import com.fasterxml.jackson.databind.node.TextNode;
import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.models.GJsonListValue;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.JsonAttrValue;
import com.unitesk.requality.api.models.JsonEmpty;
import com.unitesk.requality.api.models.JsonEnumDefValue;
import com.unitesk.requality.api.models.JsonEnumValue;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.api.models.JsonInputAttributes;
import com.unitesk.requality.api.models.JsonInputMoveArgument;
import com.unitesk.requality.api.models.JsonResponseWrapper;
import com.unitesk.requality.api.services.nodes.output.AttributeList;
import com.unitesk.requality.api.services.nodes.output.AttributeResponse;
import com.unitesk.requality.api.services.nodes.output.AttributesResponse;
import com.unitesk.requality.api.services.nodes.output.NodeResponse;
import com.unitesk.requality.api.services.project.ProjectService;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.SortingSettings;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.attribute.AppliedEnumValue;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.EnumDefinitionValue;
import com.unitesk.requality.core.attribute.EnumValue;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.tools.IPathHandler;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

@Path("/projects/{p}/nodes/{n}")
public class NodesService extends BaseService {
  public static final JsonError NODE_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "node_not_found", "Specified node does not exists");
  public static final JsonError getNodeNotFound(String s){
      return new JsonError(Status.NOT_FOUND, "node_not_found", "Node not found by path "+s);
  }
  public static final JsonError ATTRIBUTE_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "attribute_not_found", "Specified attribute does not exists");

  public static final JsonError getNodeExist(TreeNode node, String id, TreeNode par){
      return new JsonError(
          Status.BAD_REQUEST, "node_already_exists", "Node "+node.getUserVisibleName()+" with the same id "+id+" already exists in "+par.getUserVisibleName());
  }

  @GET
  @Path("/type")
  public TextNode httpGetType(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    return new TextNode(getNode(project, node).getType());
  }

  @GET
  @Path("/description")
  public TextNode httpGetDescr(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    return new TextNode(getNode(project, node).getDescription());
  }

  @PUT
  @Path("/id")
  public IDMapping httpSetId(@PathParam("p") String project, @PathParam("n") UUID node,
      String newId)
      throws JsonException {
    TreeNode nde = getNode(project, node);
    TreeNode n = nde.getWorkParent().findWorkChild(newId);
    if(n!=null&&!nde.getUUId().equals(n.getUUId()))
      throw getNodeExist(n,newId,nde.getWorkParent()).exception();
    nde.setId(newId);
    nde.commitAttributes();
    return IDMapping.fromNode(nde);
  }
  @GET
  @Path("/id")
  public IDMapping httpGetId(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    return IDMapping.fromNode(getNode(project, node));
  }
  @POST
  @Path("/view")
  public JsonEmpty httpView(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    TreeNode n = getNode(project, node);
    SelectionManager.getOne().setActiveNode(n, null, EventSource.EDITOR);
    return JsonEmpty.INSTANCE;
  }

  @GET
  @Path("/allattributes")
  public AttributeList httpListAllAttrs(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    TreeNode n = getNode(project, node);
    return new AttributeList(
        n.attrs()
            .getAllAttributeKeys().stream().collect(Collectors.toList()));
  }
  @POST
  @Path("/move")
  public NodeResponse httpMoveNode(@PathParam("p") String project, 
      @PathParam("n") UUID node,
      JsonInputMoveArgument arg)
      throws JsonException {
    TreeNode n = getNode(project, node);
    if(n == null)
      throw NODE_NOT_FOUND.exception();
    TreeNode n2 = null,par = n.getTreeDB().getNode(arg.getParent());
    if(par == null)
      throw NODE_NOT_FOUND.exception();
    if(arg.getNode()!=null){
      n2 = n.getTreeDB().getNode(arg.getNode());
      if(n2 == null)
        throw NODE_NOT_FOUND.exception();
    }
    String id = n.getId();
    Index oldIndex = n.getIndex();
    Index ind = null;
    TreeNode[]nodes = par.getSortedShownChildren(
        TreeNodeComparator.getComparator(   
            SortingSettings.init(SortingMethod.BY_INDEX, false)), null);
      //if(n2!=null){
      Index start = Index.getZero(), pstart = Index.getZero();
      boolean stop = false;
      int i;
      for(i = 0;i<nodes.length && !stop;i++){
        if(n2!=null && nodes[i].getUUId().equals(n2.getUUId()))
          stop = true;
        Index ni = nodes[i].getIndex();
        if(ni==null||ni.isUndefined()){
          int j = i+1;
          Index end = null;
          for(;j<nodes.length
              &&(nodes[j].getUUId().equals(n.getUUId())|| nodes[j].getIndex()==null||nodes[j].getIndex().isUndefined());j++);
          if(j<nodes.length&&(!nodes[j].getUUId().equals(n.getUUId())&&nodes[j].getIndex()!=null && !nodes[i].getIndex().isUndefined())){
            end = nodes[j].getIndex();
            ni = (start.findMiddleIndex(end));
          }
          else
            ni = (start.findGreaterIndex());
          nodes[i].setIndex(ni);
          nodes[i].commitAttributes();
        }
        pstart = start;
        start = ni;
      }
    //}
      boolean toEnd = n2 != null && arg.getPosition() >= 0 || n2 == null && arg.getPosition() < 0;
      if(n2!=null || toEnd)
      {
        if(toEnd){
          Index end = null;
          int j = i;
          for(;j<nodes.length
              &&(nodes[j].getUUId().equals(n.getUUId())||nodes[j].getIndex()==null||nodes[j].getIndex().isUndefined());j++);
          if(j<nodes.length&&(!nodes[j].getUUId().equals(n.getUUId())&&nodes[j].getIndex()!=null && !nodes[i].getIndex().isUndefined()))
            end = nodes[j].getIndex();
          if(end!=null)//bug
            ind = start.findMiddleIndex(end);
          else
            ind = start.findGreaterIndex();
        }
        else//add before n2
          ind = pstart.findMiddleIndex(start);//mid before prev and n2
      }
      else//add to start
        if(nodes.length>1 || nodes.length>0 && !par.getUUId().equals(n.getWorkParentUUID()))
          ind = Index.getZero().findMiddleIndex(nodes[0].getIndex());
    if((par.getQualifiedId()+"/").startsWith(n.getQualifiedId()+"/"))
      throw new RuntimeException("Cannot move node "+n.getUserVisibleName()+" to itself or its accessors("+par.getUserVisibleName()+")");
    if(!par.getUUId().equals(n.getWorkParentUUID())){
      TreeNode na = par.findWorkChild(id);
      if(na!=null&&!na.getUUId().equals(n.getUUId()))
        throw getNodeExist(na,id,par).exception();
      n.setParent(par);
      if(!n.getWorkParent().equals(par))
        throw new RuntimeException("Move was failed due to internal reasons");
    }
    if(ind != null){
      int np = -1, ip = -1;
      for(i = 0;i<nodes.length;i++){
        Index indn = nodes[i].getIndex();
        if(nodes[i].getUUId().equals(n.getUUId()))
          continue;//ignore self
        if(indn.compareTo(oldIndex)<0)
          np = i;
        if(indn.compareTo(ind)<0)
          ip = i;
      }
      if (np != ip) {
        assert ind != Index.getUndefined();
        n.setIndex(ind);
        n.commitAttributes();
      }
    }
    return new NodeResponse(par, 1);
  }

  @POST
  @Path("/move-rename")
  public NodeResponse httpMoveRenameNode(@PathParam("p") String project, 
      @PathParam("n") UUID node,
      String node2)
      throws JsonException {
    TreeNode n = getNode(project, node);
    TreeNode n2 = n.getTreeDB().getNode(node2,com.unitesk.requality.core.Path.QID|
        com.unitesk.requality.core.Path.UUID| com.unitesk.requality.core.Path.UV);
    if(n2 == null)
      throw NODE_NOT_FOUND.exception();
    String id = n.getId();
    boolean hasId = id != null && !id.isEmpty();

    if (hasId && n2.findWorkChild(id) != null) n.setId(RequalityCLI.getNextId(
        Arrays.asList(n2.getWorkChildren()),
        n.getWorkParent(), n.getType(), ""));
    n.setParent(n2);
    return new NodeResponse(n2, 1);
  }

  @GET
  @Path("/attributes")
  public AttributeList httpListAttrs(@PathParam("p") String project, @PathParam("n") UUID node)
      throws JsonException {
    TreeNode n = getNode(project, node);
    return new AttributeList(
        n.attrs()
            .getAllAttributeKeys()
            .stream()
            .filter(x -> !x.startsWith("_"))
            .collect(Collectors.toList()));
  }

  @DELETE
  @Path("/attributes")
  public AttributeList httpListAttrs(@PathParam("p") String project, @PathParam("n") UUID node,
      String attrs)
      throws JsonException {
    TreeNode n = getNode(project, node);
    for(String s:attrs.split(","))
      n.attrs().removeAttribute(s.trim());
    n.commitAttributes();
    return new AttributeList(
        n.attrs()
            .getAllAttributeKeys()
            .stream()
            .filter(x -> !x.startsWith("_"))
            .collect(Collectors.toList()));
  }
  @GET
  @Path("/attributes/{a}")
  public AttributeResponse httpGetAttr(
      @PathParam("p") String project, @PathParam("n") UUID node, @PathParam("a") String attrName)
      throws JsonException {
    TreeNode n = getNode(project, node);
    return AttributeResponse.getOne(n.attrs(), getAttribute(n, attrName),false);
  }

//  @PUT
//  @Path("/attributes/{a}/STRING")
//  public AttributeSingleResponse httpPutStringAttr(
//      @PathParam("p") String project,
//      @PathParam("n") UUID node,
//      @PathParam("a") String attrName,
//      String value)
//      throws JsonException {
//    TreeNode n = getNode(project, node);
//    n.attrs().setAttribute(new AttributeDefinition(n.getUUId(), attrName, 
//            AttributeDefinition.detectValueProvider(attrName, AttributeType.detectType(value),value)));
//    n.commitAttributes();
//    return new AttributeSingleResponse(n.attrs(), getAttribute(n, attrName));
//  }
  @PUT
  @Path("/attributes/{a}/{t}")
  public AttributeResponse httpPutTypedAttr(
      @PathParam("p") String project,
      @PathParam("n") UUID node,
      @PathParam("a") String attrName,
      @PathParam("t") String attrType,
      String value)
      throws JsonException {
    String attrType2 = attrType.toLowerCase();
    boolean found = false;
    AttributeType at = AttributeType.STRING;
    for(AttributeType at2:at.values())
      if(at2.toString().toLowerCase().equals(attrType2)){
        at = at2;
        found = true;
      }
    String en = null;
    if(!found)
      en = attrType;
    TreeNode n = getNode(project, node);
    if(attrName.equals(TreeNode.ATTR_ID))
    {
      TreeNode nd = n.getWorkParent().findWorkChild(value);
      if(nd!=null && !n.getUUId().equals(nd.getUUId()))
        throw new RuntimeException("Failed to set _id for "+n.getUserVisibleName()+". Node with _id "+value+" already exists in "+n.getWorkParent().getUserVisibleName());
    }

    n.attrs().setAttribute(new AttributeDefinition(n.getUUId(), attrName, 
            AttributeDefinition.detectValueProvider(attrName, at, 
                Attribute.convertValue(n, AttributeType.STRING, at, value, en))));
    n.commitAttributes();
    return AttributeResponse.getOne(n.attrs(), getAttribute(n, attrName),false);
  }
  @PUT
  @Path("/attributes/")
  public AttributesResponse httpPutAttributes(
      @PathParam("p") String project,
      @PathParam("n") UUID node,
      JsonInputAttributes json)
      throws JsonException {
    Set<String> changes = new HashSet<>();
    TreeNode n = getNode(project, node);
    if(n == null)
      throw getNodeNotFound(node.toString()).exception();
    ProjectService.sb.append(node+"  attrs\n");
    for(Entry<String, ? extends JsonAttrValue> en : json.getAttributes().entrySet())
    {
      ProjectService.sb.append("\t"+en.getKey()+" "+en.getValue().toString() +"\n"); 
      JsonAttrValue a = en.getValue();
      String key = en.getKey();
      AttributeType at = AttributeType.STRING;
      if(a.getType()!=null)
        at = a.getType();
      Object val = jsonAttrToValue(a, at, null, n, key);
      if(val == null)
        throw new RuntimeException("Attribute "+key+" cannot be converted for "+n.getUserVisibleName());
      if(key.equals(TreeNode.ATTR_ID))
      {
        TreeNode nd = n.getWorkParent().findWorkChild(val.toString());
        if(nd!=null && !n.getUUId().equals(nd.getUUId()))
          throw new RuntimeException("Failed to set _id for "+n.getUserVisibleName()+". Node with _id "+val.toString()+" already exists in "+n.getWorkParent().getUserVisibleName());
      }
      n.attrs().setAttribute(new AttributeDefinition(n.getUUId(), key, 
              AttributeDefinition.detectValueProvider(key, at, 
                  val)));
      changes.add(key);
    }
    if(!changes.isEmpty())
      n.commitAttributes();
    Map<String, AttributeResponse> map = new HashMap<>();
    for(String s:changes)
      map.put(s, AttributeResponse.getOne(n.attrs(), n.attrs().getAttribute(s),false));
    return new AttributesResponse(map);
  }
  @POST
  @Path("/attributes/")
  public AttributesResponse httpSetAttributes(
      @PathParam("p") String project,
      @PathParam("n") UUID node,
      JsonInputAttributes json)
      throws JsonException {
    TreeNode n = getNode(project, node);
    Set<String> torem = new HashSet<>();
    for(Entry<String, Attribute> a:n.attrs().getActiveAttributes().entrySet())
      if(a.getValue().isUserVisible())
        torem.add(a.getKey());
    if(!torem.isEmpty()){
      for(String s:torem)
        n.attrs().removeAttribute(s);
      n.commitAttributes();
    }
    return httpPutAttributes(project, node, json);
  }

  @POST
  @Path("/attributes/")
  @Consumes(MediaType.MULTIPART_FORM_DATA)
  public AttributesResponse httpFormSetAttributes(
      @PathParam("p") String project,
      @PathParam("n") UUID node,
      @FormDataParam("input") JsonInputAttributes json, 
      FormDataMultiPart mp)
      throws JsonException {
    TreeNode n = getNode(project, node);
    Set<String> torem = new HashSet<>();
    for(Entry<String, Attribute> a:n.attrs().getActiveAttributes().entrySet())
      if(a.getValue().isUserVisible())
        torem.add(a.getKey());
    if(!torem.isEmpty()){
      for(String s:torem)
        n.attrs().removeAttribute(s);
      n.commitAttributes();
    }
    return httpFormPutAttributes(project, node, json, mp);
  }
  @PUT
  @Path("/attributes/")
  @Consumes(MediaType.MULTIPART_FORM_DATA)
  public AttributesResponse httpFormPutAttributes(
      @PathParam("p") String project,
      @PathParam("n") UUID node,
      @FormDataParam("input") JsonInputAttributes json, 
      FormDataMultiPart mp)
      throws JsonException {
    Set<String> changes = new HashSet<>();
    TreeNode n = getNode(project, node);
    if(n == null)
      throw getNodeNotFound(node.toString()).exception();
    List<FormDataBodyPart> list = mp.getFields("images");
    for(Entry<String, ? extends JsonAttrValue> en : json.getAttributes().entrySet())
    {
      JsonAttrValue a = en.getValue();
      String key = en.getKey();
      AttributeType at = AttributeType.STRING;
      if(a.getType()!=null)
        at = a.getType();
      Object val = jsonAttrToValue(a, at, list, n, key);
      if(key.equals(TreeNode.ATTR_ID))
      {
        TreeNode nd = n.getWorkParent().findWorkChild(val.toString());
        if(nd!=null && !n.getUUId().equals(nd.getUUId()))
          throw new RuntimeException("Failed to set _id for "+n.getUserVisibleName()+". Node with _id "+val.toString()+" already exists in "+n.getWorkParent().getUserVisibleName());
      }

      n.attrs().setAttribute(new AttributeDefinition(n.getUUId(), key, 
              AttributeDefinition.detectValueProvider(key, at, 
                  val)));
      changes.add(key);
    }
    if(!changes.isEmpty())
      n.commitAttributes();
    Map<String, AttributeResponse> map = new HashMap<>();
    for(String s:changes)
      map.put(s, AttributeResponse.getOne(n.attrs(), n.attrs().getAttribute(s),false));
    return new AttributesResponse(map);
  }

  
  private Object jsonAttrToValue(JsonAttrValue a, AttributeType at, List<FormDataBodyPart> list, TreeNode n, String key) {
    if(at == null)
       at = a.getType();
    if(at == null)
      at = AttributeType.STRING;
    Object val = a.getValue();
    switch(at){
      case ENUM:{
        //TODO enum of htmlValues
        com.unitesk.requality.api.models.AppliedEnumValue aev = 
            (com.unitesk.requality.api.models.AppliedEnumValue) a;
        if(aev.getBaseKey() == null)
          throw new WebApplicationException("There is no enum name for attribute "+key);
        Attribute basea = n.getTreeDB().getRootNode().attrs().getAttribute(aev.getBaseKey());
        if(basea == null || basea.getType()!=AttributeType.ENUM_DEFINITION)
          throw new WebApplicationException("There is no enum definition with name "+aev.getBaseKey()+" in project");
        EnumDefinitionValue edv = (EnumDefinitionValue) basea.getValue();
        Object vala = aev.getValue();
        if(edv.getType()!=AttributeType.STRING)
          vala = Attribute.convertValue(n, AttributeType.STRING, edv.getType(), vala, null);
        for(EnumValue ev:edv){
          if(ev.getValue().equals(vala))//TODO more proper
            return new AppliedEnumValue(vala, ev.getComment(), edv.getType(), aev.getBaseKey());
        }
        throw new WebApplicationException("Enum value "+aev.getValue()+" was not found in definition with name "+aev.getBaseKey()); 
      }
      case ENUM_DEFINITION:{
        JsonEnumDefValue edv = (JsonEnumDefValue)a;
        AttributeType lt = AttributeType.STRING;
        List<EnumValue> lv = new ArrayList<>();
        if(edv.getListType()!=null)
          lt = edv.getListType();
        for(JsonAttrValue v:edv.getValues()){
          JsonEnumValue aev = (JsonEnumValue)v;
          lv.add(new EnumValue(aev.getValue(), aev.getComment()));
          //lv.add(jsonAttrToValue(a, at, list, n, key));
        }
        return new EnumDefinitionValue(lt, lv);
      }
      case LIST:{
        GJsonListValue<JsonAttrValue> alist = (GJsonListValue<JsonAttrValue>)a;
        AttributeType lt = AttributeType.STRING;
        List<Object> lv = new ArrayList<>();
        if(alist.getListType()!=null)
          lt = alist.getListType();
        for(JsonAttrValue v:alist.getValues())
          lv.add(jsonAttrToValue(v, lt, list, n, key));
        return new ListValue(lt, lv);
      }
    }
    if(val!=null && at == AttributeType.HTML){
      String os = val.toString();
      List<File> hndes = new LinkedList<>();
      if(list!=null)
        val = RequalityCLI.updateImages(URI.create(""), os, hndes,
          new IPathHandler(){
          
            @Override
            public String handlePath(URI dpath, String string, TreeNode object, List<File> added) {
              String src = CoreUtils.URLEncode(string);
              String fileName = null;
              for(FormDataBodyPart f:list){
                fileName = f.getContentDisposition().getFileName();
                if(fileName!=null&&fileName.equals(src)){
                  InputStream is = null;
                  try{
                    try {
                      is = wrapByTempFile(f.getEntityAs(InputStream.class));
                    } catch (FileNotFoundException e) {
                      e.printStackTrace();
                    }
                    n.setResourceContent(key+"/"+fileName, is);
                    return "{node.resURL}/"+fileName;
                  }finally{
                    if(is!=null)
                      try {
                        is.close();
                      } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                      }
                  }
                }
              } 
              return null;
            }
          });
    }
    return val;
  }

  private static InputStream wrapByTempFile(InputStream is) throws FileNotFoundException {
    File t = null;
    try {
      t = File.createTempFile("temp","");
      t.deleteOnExit();
    } catch (IOException e1) {
      e1.printStackTrace();
    }
    //if(fileName!=null&&fileName.equals(src)){
    try{
      int read = 0;
      byte[] bytes = new byte[1024];
      OutputStream out = new FileOutputStream(t);
      while ((read = is.read(bytes)) != -1)
      {
        out.write(bytes, 0, read);
      }
      out.flush();
      out.close();
    } catch (IOException e){
      throw new WebApplicationException("Error while uploading file. Please try again !!");
    }
    return new FileInputStream(t);
  }

  @DELETE
  @Path("/attributes/{a}")
  public JsonEmpty httpDelAttr(
      @PathParam("p") String project, @PathParam("n") UUID node, @PathParam("a") String attrName)
      throws JsonException {
    TreeNode n = getNode(project, node);
    Attribute attr = n.attrs().removeAttribute(attrName);
    if (attr != null) n.commitAttributes();
    return JsonEmpty.INSTANCE;
  }

  public static TreeNode getNode(TreeDB db, UUID uuid) throws JsonException {
    TreeNode node = db.getNode(uuid);

    if (node == null || node.isDisposed())
      throw getNodeNotFound(uuid.toString()).exception();

    return node;
  }

  public static TreeNode getNode(String project, UUID uuid) throws JsonException {
    return getNode(ProjectService.getProject(project), uuid);
  }

  public static Attribute getAttribute(TreeNode node, String attrName) throws JsonException {
    Attribute attr = node.attrs().getAttribute(attrName);

    if (attr == null) throw ATTRIBUTE_NOT_FOUND.exception();

    return attr;
  }
}
