package com.unitesk.requality.api.services.nodes.output;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.unitesk.requality.api.models.JsonOutputBase;
import com.unitesk.requality.core.attribute.AppliedEnumValue;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.EnumDefinitionValue;
import com.unitesk.requality.core.attribute.EnumValue;
import com.unitesk.requality.core.attribute.ListValue;

public abstract class AttributeResponse extends JsonOutputBase {
  protected AttributeStorage storage;
  protected AttributeType at;
  protected Object val;
  public AttributeResponse(AttributeStorage storage, AttributeType at, Object val) {
    this.storage = storage;
    this.val = val;
    this.at = at;

  }
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public String getType() {
    if(at == null)
      return null;
    return at.name();
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((val == null) ? 0 : val.hashCode());
    result = prime * result + ((at == null) ? 0 : at.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    AttributeSingleResponse other = (AttributeSingleResponse) obj;
    if (at == null) {
      if (other.at != null) return false;
    } else if (!at.equals(other.at)) return false;
    if (val == null) {
      if (other.val != null) return false;
    } else if (!val.equals(other.val)) return false;
    return true;
  }

  @Override
  public String toString() {
    return "AttributeResponse [attr=" + at + "-"+val+"]";
  }

  public static AttributeResponse getOne(AttributeStorage attrs, Attribute attribute,boolean list) {
    if(attribute == null)
      return null;
    return getOne(attrs, attribute.getType(), attribute.getValue(),list);
  }

  public static AttributeResponse getOne(AttributeStorage storage, AttributeType type, Object o,boolean list) {
    if(o.getClass().equals(EnumValue.class)){
      EnumValue ev = (EnumValue)o;
      return new EnumValueResponse(storage, list?null:type,ev.getValue(),ev.getComment());
    }
    if(type == AttributeType.ENUM_DEFINITION)
    {
      EnumDefinitionValue ed = (EnumDefinitionValue) o;
      return new AttributeEnumDefResponse(storage, list?null:type,ed.getType(), ed);
    }
    if(type == AttributeType.ENUM){
      AppliedEnumValue aev = (AppliedEnumValue)o;
      return new AppliedEnumResponse(storage, list?null:type, aev.getType(),aev.getValue(),aev.getComment(),aev.getBaseKey());
    }
    if(type == AttributeType.REFERENCE)
      return (new ReferenceResponse(storage, list?null:AttributeType.REFERENCE, o, true));
    else
    if(type == AttributeType.LIST){
      ListValue lv = (ListValue)o;
      return (new AttributeListResponse(storage, list?null:AttributeType.LIST,lv.getType(), o));
    }else
      return (new AttributeSingleResponse(storage, list?null:type, o));
  }
}