package com.unitesk.requality.api.services.nodes.output;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import com.unitesk.requality.api.models.JsonOutputBase;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.VirtualNode;

public class NodeResponse extends JsonOutputBase {
  private String id;
  private UUID uid,parentuuid;
  private String uvid;
  private String nid;
  private AttributesResponse attrs;
  private Map<String, NodeResponse> childs = null;
  
  public String getId(){
    return id;
  }
  public String getNameOrId(){
    return nid;
  }
  public UUID getParent(){
    return parentuuid;
  }
  public UUID getUUID(){
    return uid;
  }

  public String getUserVisibleId(){
    return uvid;
  }
  public Map<String, AttributeResponse> getAttributes(){
    return attrs.getAttributes();
  }
  public Map<String, NodeResponse> getChildren(){
    return childs;
  }
  public NodeResponse(TreeNode node, int depth) {
    id = node.getId();
    uid = node.getUUId();
    nid = node.getNameOrId();
    uvid = node.getUserVisibleName();
    parentuuid = node.getShownParentUUID();
    this.attrs = new AttributesResponse(node);
    if(depth>0||depth == -1){
      childs = new HashMap<>();
      for(TreeNode ch: node.getSortedShownChildren()){
        String id = ch.getId();
        if(ch.getWorkParent().getType().equals(VirtualNode.TYPE_NAME))
          id = formVirtId(ch);
        childs.put(id,new NodeResponse(ch, depth!=-1?depth-1:-1));//TODO mult id
      }
    }
  }

  private String formVirtId(TreeNode ch) {
    TreeNode wp = ch.getWorkParent();
    String id = ch.getId();
    if(wp == null)
      return id;
    if(wp.getType().equals(VirtualNode.TYPE_NAME))
      return formVirtId(wp)+"/"+id;
    return id; 
  }
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((id == null) ? 0 : id.hashCode());
    result = prime * result + ((uid == null) ? 0 : uid.hashCode());
    result = prime * result + ((attrs == null) ? 0 : attrs.hashCode());
    result = prime * result + ((childs == null) ? 0 : childs.hashCode());
    result = prime * result + ((parentuuid == null) ? 0 : parentuuid.hashCode());
    result = prime * result + ((nid == null) ? 0 : nid.hashCode());
    result = prime * result + ((uvid == null) ? 0 : uvid.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    NodeResponse other = (NodeResponse) obj;
    if (id == null) {
      if (other.id != null) return false;
    } else if (!id.equals(other.id)) return false;
    if (uid == null) {
      if (other.uid != null) return false;
    } else if (!uid.equals(other.uid)) return false;
    if (attrs == null) {
      if (other.attrs != null) return false;
    } else if (!attrs.equals(other.attrs)) return false;
    if (childs == null) {
      if (other.childs != null) return false;
    } else if (!childs.equals(other.childs)) return false;
    if (nid == null) {
      if (other.nid != null) return false;
    } else if (!nid.equals(other.nid)) return false;
    if (parentuuid == null) {
      if (other.parentuuid != null) return false;
    } else if (!parentuuid.equals(other.parentuuid)) return false;
    if (uvid == null) {
      if (other.uvid != null) return false;
    } else if (!uvid.equals(other.uvid)) return false;
    return true;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("NodeResponse [id=");
    sb.append(id);
    sb.append(",uuid=");
    sb.append(uid);
    sb.append(",parent=");
    sb.append(parentuuid);
    sb.append(", attributes=");
    sb.append(attrs);
    if(childs!=null){
      sb.append("children=");
      sb.append(childs);
    }
    sb.append("]");
    return sb.toString();
  }
}
