package com.unitesk.requality.api.services.project;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response.Status;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.JsonEmpty;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.api.models.JsonInputAttributes;
import com.unitesk.requality.api.services.nodes.NodesService;
import com.unitesk.requality.api.services.nodes.output.AttributeEnumDefResponse;
import com.unitesk.requality.api.services.nodes.output.AttributeResponse;
import com.unitesk.requality.api.services.nodes.output.AttributesResponse;
import com.unitesk.requality.api.services.nodes.output.NodeResponse;
import com.unitesk.requality.api.services.project.input.QIDRequest;
import com.unitesk.requality.api.services.project.output.ProjectList;
import com.unitesk.requality.api.services.project.output.ProjectResponse;
import com.unitesk.requality.api.services.project.output.QIDResponse;
import com.unitesk.requality.api.services.requirement.input.CreateRequirement;
import com.unitesk.requality.api.services.requirement.input.ImportHtml;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.documents.OmnivorousImporter;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.eclipse.views.documents.UpdateFromHTMLModule;
import com.unitesk.requality.eclipse.wizards.NewRequalityProjectWizard;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.RequalityCLI;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.FormDataParam;

@Path("/projects")
public class ProjectService extends BaseService {
  public static final JsonError PARENT_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "parent_not_found", "Parent node does not exists");

  public static final JsonError PROJECT_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "project_not_found", "Specified project does not exists");
  public static final JsonError PROJECT_EXISTS =
      new JsonError(Status.BAD_REQUEST, "project_exists", "Specified project already exists");
  public static final JsonError QID_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "qid_not_found", "Specified QID does not exists");

  private EclipseTracker trees = EclipseTracker.getTracker();

  @GET
  public ProjectList httpList() {
    return new ProjectList(
        Stream.of(trees.getTrees()).map(TreeDB::getName).collect(Collectors.toList()));
  }
  private String getFixedHeaderValue(String val) {
    String result = val;
    try {
      result = new String(result.getBytes(), "utf8");
    } catch (UnsupportedEncodingException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return result;
}
 

  // save uploaded file to new location
  private void writeToFile(InputStream uploadedInputStream,
      String uploadedFileLocation) {
    OutputStream out = null;
    try {
      File to = new File(
          uploadedFileLocation);
      if(!to.getParentFile().exists())
        to.getParentFile().mkdirs();
      to.createNewFile();
      int read = 0;
      byte[] bytes = new byte[1024];
      out = new FileOutputStream(new File(uploadedFileLocation));
      while ((read = uploadedInputStream.read(bytes)) != -1) {
          out.write(bytes, 0, read);
      }
    } catch (IOException e) {

        e.printStackTrace();
    }
    finally{
      if(out != null){
        try {
          out.flush();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        try {
          out.close();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        
      }
      if(out!=null){
        try {
          out.close();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
  }
  @GET
  @Path("/{project}")
  public ProjectResponse httpGetProject(@PathParam("project") String project) throws JsonException {
    TreeDB db = getProject(project);
    return new ProjectResponse(db.getName(), db.getRootNode());
  }

  @PUT
  @Path("/{project}")
  public JsonEmpty httpCreateProject(@PathParam("project") String project) throws JsonException {
    for (TreeDB db : trees.getTrees()) {
      if (db.getName().equals(project)) {
        throw PROJECT_EXISTS.exception();
      }
    }

    TreeDB fresh = NewRequalityProjectWizard.createProject(project);
    if (fresh != null) return JsonEmpty.INSTANCE;
    else throw JsonError.INTERNAL_ERROR.exception();
  }

  @DELETE
  @Path("/{project}")
  public JsonEmpty httpDeleteProject(@PathParam("project") String project) throws JsonException {
    TreeDB db = getProject(project);
    IProject proj = trees.getProject(db);
    if (proj != null) {
      try {
        proj.refreshLocal(IResource.DEPTH_INFINITE, null);
        proj.delete(true, null);
        return JsonEmpty.INSTANCE;
      } catch (CoreException e) {
        e.printStackTrace();
        throw JsonError.INTERNAL_ERROR.exception();
      }
    } else {
      System.err.println("While deleting project '" + project + "': trees.getProject() == null");
      throw JsonError.INTERNAL_ERROR.exception();
    }
  }

  @GET
  @Path("/{project}/tree-{d}")
  public List<NodeResponse> httpReqTreeD(@PathParam("project") String project,
      @PathParam("d") String d)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode nde = db.getNode("/");
    if(nde == null)
      throw NodesService.NODE_NOT_FOUND.exception();
    List<NodeResponse>nodes = new LinkedList<>();
    int depth = -1;
    try{
      depth = Integer.parseInt(d);
    }
    catch(Exception e){
    }
    for(TreeNode node:nde.getShownChildren())
      nodes.add(new NodeResponse(node, depth));
    return nodes;
  }
  @POST
  @Path("/{project}/tree-{d}")
  public NodeResponse httpReqTreePostD(@PathParam("project") String project,
      @PathParam("d") String d,
      String node)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode nde = db.getNode(node);
    if(nde == null)
      throw NodesService.NODE_NOT_FOUND.exception();
    int depth = -1;
    try{
      depth = Integer.parseInt(d);
    }
    catch(Exception e){
    }
    return new NodeResponse(nde, depth);
  }
  @GET
  @Path("/{project}/tree")
  public NodeResponse httpReqTree(@PathParam("project") String project)
      throws JsonException {
    return httpReqTreePost(project,"/Requirements");
  }
  @POST
  @Path("/{project}/tree")
  public NodeResponse httpReqTreePost(@PathParam("project") String project
      ,String node)
      throws JsonException {
    return httpReqTreePostD(project, "-1", node);
//    TreeDB db = getProject(project);
//    TreeNode nde = db.getNode(node);
//    if(nde == null)
//      throw NodesService.NODE_NOT_FOUND.exception();
//    return new NodeResponse(nde,-1);
  }
  @GET
  @Path("/{project}/tree/{node}/")
  public NodeResponse httpReqTreeGET(@PathParam("project") String project
      , @PathParam("node") String node)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode nde = db.getNode(UUID.fromString(node));
    if(nde == null)
      throw NodesService.NODE_NOT_FOUND.exception();
    return new NodeResponse(nde, -1);
  }
  @POST
  @Path("/{project}/node")
  public NodeResponse getNode(@PathParam("project") String project,
      String qid)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getNode(qid,com.unitesk.requality.core.Path.UV|
        com.unitesk.requality.core.Path.QID|com.unitesk.requality.core.Path.UUID);
    if (node == null) {
      throw QID_NOT_FOUND.exception();
    } else {
      return new NodeResponse(node,0);
    }
  }
  @GET
  @Path("/{project}/enumDefs")
  public AttributesResponse getEDefs(@PathParam("project") String project)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getRootNode();
    if (node == null) { 
      throw QID_NOT_FOUND.exception();
    } else {
      java.util.Map<String,AttributeResponse> map = new HashMap<>();
      for(Entry<String, Attribute> en:node.attrs().getActiveAttributes().entrySet()){
        Attribute a = en.getValue();
        if(a.getType()==AttributeType.ENUM_DEFINITION){
          map.put(en.getKey(), new AttributeEnumDefResponse(node.attrs(), a.getType(),a.getListType(), a.getValue()));
        }
      }
      return new AttributesResponse(map);
    }
  }
  @POST
  @Path("/{project}/enumDefs")
  public AttributesResponse setEDefs(@PathParam("project") String project
      , JsonInputAttributes atrs)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getRootNode();
    if (node == null) { 
      throw QID_NOT_FOUND.exception();
    } else {
      new NodesService().httpPutAttributes(project, node.getUUId(), atrs);
      Set<String> changes = new HashSet<>();
      for(Entry<String, Attribute> en:node.attrs().getActiveAttributes().entrySet())
        if(!atrs.getAttributes().containsKey(en.getKey()) && en.getValue().getType() == AttributeType.ENUM_DEFINITION){
          changes.add(en.getKey());
        }
      if(!changes.isEmpty()){
        for(String s:changes)
          node.attrs().removeAttribute(s);
        node.commitAttributes();
      }
      java.util.Map<String,AttributeResponse> map = new HashMap<>();
      for(Entry<String, Attribute> en:node.attrs().getActiveAttributes().entrySet()){
        Attribute a = en.getValue();
        if(a.getType()==AttributeType.ENUM_DEFINITION){
          map.put(en.getKey(), new AttributeEnumDefResponse(node.attrs(), a.getType(),a.getListType(), a.getValue()));
        }
      }
      return new AttributesResponse(map);
    }
  }
  @PUT
  @Path("/{project}/enumDefs")
  public AttributesResponse putEDefs(@PathParam("project") String project
      , JsonInputAttributes atrs)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getRootNode();
    if (node == null) { 
      throw QID_NOT_FOUND.exception();
    } else {
      new NodesService().httpPutAttributes(project, node.getUUId(), atrs);
      java.util.Map<String,AttributeResponse> map = new HashMap<>();
      for(Entry<String, Attribute> en:node.attrs().getActiveAttributes().entrySet()){
        Attribute a = en.getValue();
        if(a.getType()==AttributeType.ENUM_DEFINITION){
          map.put(en.getKey(), new AttributeEnumDefResponse(node.attrs(), a.getType(),a.getListType(), a.getValue()));
        }
      }
      return new AttributesResponse(map);
    }
  }
  @DELETE
  @Path("/{project}/node")
  public NodeResponse deleteNode(@PathParam("project") String project,
      String qid)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getNode(qid,com.unitesk.requality.core.Path.UV|
        com.unitesk.requality.core.Path.QID|com.unitesk.requality.core.Path.UUID);
    if (node == null) {
      throw QID_NOT_FOUND.exception();
    } else {
      TreeNode sp = node.getShownParent();
      node.delete();
      if(sp!=null)
        return new NodeResponse(sp,1);
    }
    return null;
  }
  @POST
  @Path("/{project}/qid-resolve")
  public QIDResponse httpQidResolve(@PathParam("project") String project, QIDRequest qid)
      throws JsonException {
    TreeDB db = getProject(project);
    TreeNode node = db.getNode(qid.getQid());
    if (node == null) {
      throw QID_NOT_FOUND.exception();
    } else {
      return new QIDResponse(IDMapping.fromNode(node), node.getType());
    }
  }

  @Path("{p}/create")
  @PUT
  @Consumes(MediaType.MULTIPART_FORM_DATA)
  public NodeResponse httpFormCreate(@PathParam("p") String project,
      @FormDataParam("input")CreateRequirement input, 
      FormDataMultiPart mp)
      throws JsonException {
    try{
      TreeDB db = getProject(project);
      TreeNode parent = db.getNode(input.getParent());
      if (parent == null) throw PARENT_NOT_FOUND.exception();
  
      String id = input.getId();
      boolean hasId = id != null && !id.isEmpty();
      TreeNode n = null;
      if (hasId){
        n = parent.findWorkChild(id);
        if(n != null)
         throw NodesService.getNodeExist(n,id, parent).exception();
      }
  
      if (!hasId) id = RequalityCLI.getNextId(parent, input.getType(), null);
  
      TreeNode child = parent.createChild(id, input.getType());
      NodesService ns = new NodesService();
      if (child == null) throw JsonError.INTERNAL_ERROR.exception();
      try{
        ns.httpFormPutAttributes(project, child.getUUId(), input, mp);
        return new NodeResponse(child, 0);
      }catch(Exception e){
        child.delete();
        throw e;
      }
    }
    catch(Exception e)
    {
      throw new WebApplicationException(CoreUtils.getTrace(e));
    }
  }
  public static StringBuilder sb = new StringBuilder();
  @Path("{p}/create")
  @PUT
  public NodeResponse httpCreate(@PathParam("p") String project, CreateRequirement input)
      throws JsonException {

    try{
      TreeDB db = getProject(project);
      TreeNode parent = db.getNode(input.getParent());
      if (parent == null) throw PARENT_NOT_FOUND.exception();
  
      String id = input.getId();
      boolean hasId = id != null && !id.isEmpty();
      TreeNode n = null;
      if (hasId){
        n = parent.findWorkChild(id);
        if(n != null)
         throw NodesService.getNodeExist(n,id, parent).exception();
      }
  
  
      if (!hasId) id = RequalityCLI.getNextId(parent, input.getType(), null);
  
      TreeNode child = parent.createChild(id, input.getType());
      NodesService ns = new NodesService();

      sb.append("~~~\n");
      if(child!=null)
        sb.append(child.getUUId()+"  create\n");
      if (child == null) throw JsonError.INTERNAL_ERROR.exception();
      try{ 
        ns.httpPutAttributes(project, child.getUUId(), input);
      }catch(Exception e){
        child.delete();
        throw e;
      }
      if(CoreUtils.DEBUG_LEVEL>0){
        System.out.println(sb.toString());
        sb.setLength(0);
      }
      return new NodeResponse(child, 0);
    }
    catch(Exception e){
      throw new WebApplicationException(CoreUtils.getTrace(e));
    }
  }
  @POST
  @Path("{p}/import")
  public JsonEmpty httpImport(@PathParam("p") String project, ImportHtml input) throws Exception {
    TreeDB db = getProject(project);
    TreeNode r =  db.getNode(Requirement.getTypeRootQId());

    String cs = OmnivorousImporter.getCharset(input.getHtml());

    File inFile = File.createTempFile("requality_in_", ".xhtml");

    try (Writer w = new OutputStreamWriter(new FileOutputStream(inFile), Charset.forName(cs))) {
      w.write(input.getHtml());
      w.close();
    }
    DocFolder df = (DocFolder) db.getNode(Document.getTypeRootQId());
    String filename = inFile.getName();
    filename = filename.substring(0, filename.lastIndexOf(".")) + ".xhtml";
    int counter;
    Document doc = (Document) df.createChild(inFile.getName(), Document.TYPE_NAME);
    doc.setResourceName(inFile.getName());
    doc.commitAttributes();

    ResourceStorage rs = EclipseTracker.getResourceStorage(doc.getTreeDB());
    IFolder resFolder = rs.getResourceFolder(doc.getUUId());
    if (!resFolder.exists()) {
      resFolder.create(true, true, null);
    }

    File outFile = new File(resFolder.getLocation().toOSString(), filename);

    if (!outFile.exists() || outFile.equals(inFile)) {
      OmnivorousImporter importer = new OmnivorousImporter();

      String docId = doc.getQualifiedId();
      importer.importDocument(
          docId,
          inFile,
          outFile,
          df.getTreeDB(),
              df.getTreeDB()
                  .getNode(Requirement.getTypeRootQId(), com.unitesk.requality.core.Path.QID),
          System.out);
    }

    UpdateFromHTMLModule up = new UpdateFromHTMLModule();
    up.setShowMessages(false);

    up.configure(doc, Collections.emptyMap(), System.out);//TODO add params?

    db.startChangeRow();
    try{
      up.process(doc.getQualifiedId(), r, System.out, null);
      doc.delete();
    }
    finally{
      db.stopChangeRow();
    }

    return JsonEmpty.INSTANCE;
  }

  /////////////////

  public static TreeDB getProject(String name) throws JsonException {
    for (TreeDB db : EclipseTracker.getTracker().getTrees()) {
      if (db.getName().equals(name)) {
        return db;
      }
    }

    throw PROJECT_NOT_FOUND.exception();
  }
}
