package com.unitesk.requality.api.services.report;

import static com.unitesk.requality.api.services.project.ProjectService.getProject;

import java.util.UUID;
import java.util.stream.Collectors;

import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.core.Response.Status;

import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.Utils;
import com.unitesk.requality.api.models.IDList;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.api.services.report.output.GeneratedReport;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.operations.ReportGenResult;
import com.unitesk.requality.eclipse.handlers.report.GenerateReportHandler;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

@Path("/projects/{p}/reportSettings")
public class ReportSettingsService extends BaseService {
  public static final JsonError REPORT_SETTINGS_NOT_FOUND =
      new JsonError(
          Status.NOT_FOUND,
          "report_settings_not_found",
          "Specified report settings does not exists");

  @GET
  public IDList httpList(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    TreeNode docRoot = db.getNode("/Reports");
    return new IDList(
        Utils.stream(new DeepFirstTreeWalker(docRoot, false, true, true).iterator())
            .filter(x -> x.getType().equals(ReportSettings.TYPE_NAME))
            .map(IDMapping::fromNode)
            .collect(Collectors.toList()));
  }

  @POST
  @Path("/{r}/generate")
  public GeneratedReport httpGen(@PathParam("p") String project, @PathParam("r") UUID repId)
      throws JsonException {
    ReportSettings setts = getReportSettings(project, repId);
    GenerateReportHandler grh = new GenerateReportHandler();
    ReportGenResult r = grh.generateReport(setts, null);//TODO work?
    RLogger.logInfo("returned from generate: '" + r.href + "'");
    return null;
  }

  public static ReportSettings getReportSettings(TreeDB db, UUID uuid) throws JsonException {
    TreeNode node = db.getNode(uuid);

    if (node == null || !node.getType().equals(ReportSettings.TYPE_NAME))
      throw REPORT_SETTINGS_NOT_FOUND.exception();

    return (ReportSettings) node;
  }

  public static ReportSettings getReportSettings(String project, UUID repId) throws JsonException {
    return getReportSettings(getProject(project), repId);
  }
}
