package com.unitesk.requality.api.services.requirement;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.core.Response.Status;

import com.fasterxml.jackson.databind.node.TextNode;
import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.Utils;
import com.unitesk.requality.api.models.IDList;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.IDTree;
import com.unitesk.requality.api.models.JsonEmpty;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.api.services.requirement.input.CreateRequirement;
import com.unitesk.requality.api.services.requirement.input.ImportHtml;

import static com.unitesk.requality.api.services.project.ProjectService.getProject;

import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.OmnivorousImporter;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.views.documents.UpdateFromHTMLModule;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.report.ReportGenerator;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

@Path("/projects/{p}/requirements")
public class RequirementService extends BaseService {
  public static final JsonError PARENT_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "parent_not_found", "Parent node does not exists");

  public static final JsonError REQUIREMENT_EXISTS =
      new JsonError(
          Status.BAD_REQUEST, "requirement_already_exists", "Specified requirement already exists");

  public static final JsonError REQUIREMENT_NOT_FOUND =
      new JsonError(
          Status.NOT_FOUND, "requirement_not_found", "Specified requirement does not exists");

  @GET
  public IDList httpList(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    TreeNode reqRoot = db.getNode(Requirement.getTypeRootQId());
    return IDList.fromIterator(new DeepFirstTreeWalker(reqRoot, false, true, true).iterator());
  }

  @GET
  @Path("/tree")
  public IDTree httpTree(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    return IDTree.fromNode(db.getNode(Requirement.getTypeRootQId()));
  }

  @GET
  @Path("/root")
  public IDMapping httpRoot(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    return IDMapping.fromNode(db.getNode(Requirement.getTypeRootQId()));
  }

  @DELETE
  @Path("/{r}")
  public JsonEmpty httpDelete(@PathParam("p") String project, @PathParam("r") UUID req)
      throws JsonException {
    TreeDB db = getProject(project);
    getRequirement(db, req).delete();
    return JsonEmpty.INSTANCE;
  }

  @GET
  @Path("/{r}/export")
  public TextNode httpExport(@PathParam("p") String project, @PathParam("r") UUID req)
      throws Exception {
    TreeDB db = getProject(project);
    Requirement r = getRequirement(db, req);

    File resRoot =
        new File(
            EclipseTracker.getResourceStorage(db)
                .getResourceFolder(db.getRootNode().getUUId())
                .getLocationURI()
                .getPath());

    File reportRoot = new File(resRoot, "tempDir/" + System.currentTimeMillis());

    if (reportRoot.exists()) Utils.deleteDirectory(reportRoot);

    if (!reportRoot.exists() && !reportRoot.mkdirs())
      throw new IOException("Failed to mkdirs " + reportRoot);

    Template t =
        Activator.getDefault()
            .getReportTemplates()
            .get("com.unitesk.requality.reports.reqSpecification");
    for (TemplateFile tf : t.getTemplates()) {
      File tff =
          new File(reportRoot, tf.getOutputFileName(db));
      if (tff.exists() && !tff.delete()) throw new IOException("Failed to delete " + tff);
    }

    Map<String, Object> userData = new HashMap<String, Object>();
    ReportGenerator.generateReport(r, t, reportRoot.toURI().toString(), userData);

    File ret =
        new File(
            reportRoot, t.getTemplates().get(0).getOutputFileName(db));
    String content = new String(Files.readAllBytes(ret.toPath()), StandardCharsets.UTF_8);
    Utils.deleteDirectory(reportRoot);

    return new TextNode(content);
  }

  public static Requirement getRequirement(TreeDB db, UUID uuid) throws JsonException {
    TreeNode node = db.getNode(uuid);

    if (node == null || !node.getType().equals(Requirement.TYPE_NAME))
      throw REQUIREMENT_NOT_FOUND.exception();

    return (Requirement) node;
  }
}
